//	GetPass 7.1 - C_FILTER for F6FBB BBS Software
//  	Written by Marco Savegnago IW3FQG

#include 	"gpmsg.h"
#include 	"gpmd2.h"
#include 	"gppass.h"
#include	 	"strhlp.h"
#include		"sap.h"
#ifdef	__INC_RSA
#include 	"radix64.h"
#endif

#include 	"gpapps.h"

#include		<dir.h>
#include 	<stdio.h>

char	Password::sz_NOMD2[]	=	"__NOMD2__";
char	Password::sz_NOMTX[]	=	"__NOMTX__";
char	Password::sz_NORSA[]	=	"__NORSA__";

Password::Password( const char *cpsz_Callsign, const char* cpsz_PasswordFilename)
:
	psz_callsign(NULL),
	passwordString(NULL),
	w_passwordFlags(PT_MTX|PT_MD2|PT_RSA),
	psz_challenge(NULL),
	b_rsaFound(false),
	b_stdFound(false)
{
#ifdef	__INC_RSA
	privateKey.bits	=	MIN_RSA_MODULUS_BITS;
#endif
	strncpy( sz_passwordFilename, cpsz_PasswordFilename, sizeof(sz_passwordFilename));
	psz_callsign	=	strdup(cpsz_Callsign);
}

Password::~Password()
{
	if(passwordString!=NULL)
		delete	passwordString;

	if(psz_callsign!=NULL)
		free(psz_callsign);

	if(psz_challenge!=NULL)
		free(psz_challenge);
}

int Password::ReadPassword()
{
	// return:
	//	0 password OK
	// 1 password not found
	// 2 I/O error

	//	wPasswordFlags:

	//  0x0001 User support Matrix
	//  0x0002 User support MD2
	//	 0x0004 User support RSA

	int i_ret = 1;

	ifstream pwd_file( sz_passwordFilename);

	if ( ! pwd_file.good() )
	{
		return( 2 );
	}

	char sz_temp[ MAXLINELEN ];
	char sz_label[ MAXLINELEN ];

	while ( pwd_file.getline( sz_temp, sizeof ( sz_temp ) ), ! pwd_file.eof() )
	{
		char *s = sz_temp;
		char *d = sz_label;

		if(isComment(s))
					continue;

		skipSpace(s);

		if(isComment(s) || *s == EOS)
					continue;

		while ( !isspace(*s) )
					*d++ = *s++;

		*d = EOS;

		if( stringMatch( psz_callsign, sz_label) )
		{
			bool	done	=	false;

			while(!done)
			{
				skipSpace(s);

				if(!strnicmp(sz_NOMD2, s, sizeof(sz_NOMD2) - 1))
				{
					w_passwordFlags	&=	~PT_MD2;
					s += sizeof(sz_NOMD2) - 1;  // 9 == len of __NOMD2__
				}
				else if(!strnicmp(sz_NOMTX, s, sizeof(sz_NOMTX) - 1))
				{
					w_passwordFlags	&=	~PT_MTX;
					s += sizeof(sz_NOMTX) - 1;  // 9 == len of __NOMTX__
				}
				else if(!strnicmp(sz_NORSA, s, sizeof(sz_NORSA) - 1))
				{
					w_passwordFlags	&=	~PT_RSA;
					s += sizeof(sz_NORSA) - 1;  // 0 == len of __NORSA__
				}
				else
				{
					char	cdelimiter = '\"';
					const char	cdefDelimiter = '\"';
					skipSpace(s);

					cdelimiter	=	(*s==cdefDelimiter) ? cdefDelimiter : ' ';
					s=GetDelimitedString(s, sz_label, sizeof(sz_label)-1, cdelimiter, cdelimiter, '\\');

					psz_challenge = strdup(sz_label);

					skipSpace(s);

					if(!checkRSAHeader(s))
					{
						cdelimiter	=	(*s==cdefDelimiter) ? cdefDelimiter : ' ';
						s=GetDelimitedString(s, sz_label, sizeof(sz_label)-1, cdelimiter, cdelimiter, '\\');

						passwordString = new	PasswordString(sz_label);
						b_stdFound = true;
					}

					skipSpace(s);
#ifdef	__INC_RSA
					if(*s && readPrivateKey(s, &privateKey))
						b_rsaFound=true;
#endif
					i_ret = 0;
					done = true;
				}
			}
			break;
		}
	}

	pwd_file.close(); // dummy

	return( i_ret );
}

int	Password::checkRSAHeader(char* psz_EncodedBlockIn)
{
	register	char*	pc_Begin;
	int					i_bitNumber;

	if((*psz_EncodedBlockIn != '[') || strncmp(++psz_EncodedBlockIn, sz_SAPHeader, strlen(sz_SAPHeader)))
		return 0;

	if(NULL!=(pc_Begin=strchr(psz_EncodedBlockIn, '-')))
	{
		pc_Begin++;

		if(*pc_Begin!=SAPMajorVersion)
			return 0;

		while(*pc_Begin&& *pc_Begin!='.')
			pc_Begin++;

		pc_Begin++;

		if(*pc_Begin!=SAPMinorVersion)
			return 0;

		if(NULL!=(pc_Begin=strchr(pc_Begin, '-')))
		{
			while(!isdigit(*pc_Begin))
				pc_Begin++;

			i_bitNumber=atoi(pc_Begin);

#ifdef	__INC_RSA
			if( i_bitNumber < MIN_RSA_MODULUS_BITS || i_bitNumber > MAX_RSA_MODULUS_BITS )
				return 0;
#endif

			return 1;
		}
	}

	return 0;
}

#ifdef	__INC_RSA
int	Password::readPrivateKey(char* psz_EncodedBlockIn, R_RSA_PRIVATE_KEY* p_privateKey)
{
	unsigned int	decodedBlockLen;
	RADIX64			radix;
	char*			pc_Begin;
	char*			pc_End;
	unsigned char*	pby_buffer;

	//	Read the header

	if(!checkRSAHeader(psz_EncodedBlockIn))
		return 0;

	//	Escape the header and get the modulus

	if(NULL!=(pc_Begin	=	strrchr(psz_EncodedBlockIn, '-')))
	{
		pc_Begin++;

		if(NULL!=(pc_End=strchr(pc_Begin, ']')))
		{
			pby_buffer	=	(unsigned char*)p_privateKey;

			if(ID_OK!=(radix.DecodeBlock(pby_buffer, &decodedBlockLen, (unsigned char*)pc_Begin, (unsigned int)(pc_End-pc_Begin))))
				return 0;
		}

      return 1;
	}

	return 0;
}
#endif

int Password::MakeReply(char* psz_Challenge, ostream&	ostream)
{
	// return:
	//		0 password OK
	// 	1 password not found
	// 	2 I/O error

	//	wPasswordFlags:

	//  0x0004 User not support MD2
	//  0x0008 User not support Matrix
	//	 0x0010 User not support RSA

	int i_ret = 1;

	char* s;
	char* d;
	char sz_temp[ MAXLINELEN];
	char sz_out[ MAXLINELEN];
	unsigned int  j=0;

	if(!b_rsaFound && !b_stdFound)
		return 1;

	if(NULL != (d=striinc(psz_challenge, psz_Challenge)))
	{
		if ( ((w_passwordFlags&PT_MD2) || (b_rsaFound && w_passwordFlags&PT_RSA)) &&  (NULL != strrchr(psz_Challenge, ']')) && (NULL != (s=strchr(psz_Challenge, '[')) ))
		{
			while(!isdigit(*s))
				 s++;

			strcpy(sz_temp,s);
			s=strchr(sz_temp,']');

			while(!isalnum(*s))
				s--;

			*(s+1)=EOS;

			bool	b_found = false;
#ifdef	__INC_RSA

			if((w_passwordFlags&PT_RSA) && b_rsaFound)
			{
				MD2 			md(sz_temp);
				RSA     		rsa;
				RADIX64 		radix;
				unsigned int 	jj;
				unsigned int	len;

				len = (privateKey.bits + 7) / 8;

				if((md.sizeofDigest() + 11) > len)
					len	-= 11;
				else
					len =	md.sizeofDigest();

				if(ID_OK==rsa.PrivateEncrypt((unsigned char*)sz_temp, &j, md.getDigest(), len, &privateKey))
				{
					if(ID_OK==radix.EncodeBlock((unsigned char*)sz_out, &jj, (unsigned char*)sz_temp, j))
					{
						sz_out[jj]=EOS;
						sprintf(sz_temp, "[%s-%c.%c-RSA%d-%s]", sz_SAPHeader, SAPMajorVersion, SAPMinorVersion, privateKey.bits, sz_out);
						j=strlen(sz_temp);
						b_found=true;
					}
				}
			}
#endif
			if(b_stdFound && (w_passwordFlags&PT_MD2) && b_found==false)
			{
				strcat(sz_temp, passwordString->getString());

				MD2 md(sz_temp);
				strcpy(sz_temp, md.getString());

				j=strlen(sz_temp);

				i_ret = 0;
			}
		}
		else if(b_stdFound && (w_passwordFlags&PT_MTX))
		{
			d=striinc(psz_challenge, psz_Challenge);
			d+=strlen(psz_challenge);

			skipSpace(d);

			if(isdigit(*d))
			{
				j=0;

				while(true)
				{
					s=d;

					if(!*s)
						break;

					skipSpace(s);

					if(!isdigit(*s))
						break;

					d=s;

					while(isdigit(*d))
						  d++;

					*d++=NULL;

					if(atoi(s) > passwordString->lenOfPassword())
						return	1;

					sz_temp[j++] = *(passwordString->getPos(atoi(s))-1);

					i_ret = 0;
				}
			}
		}
	}
	else
		i_ret = 1;

	sz_temp[j++] = 0x0d;
	sz_temp[j]=EOS;
	ostream	<< sz_temp;

	return( i_ret );
}

